"""Tool for asking human input."""


class HumanInputRun:
    """Tool that asks user for input."""

    def __init__(self, prompt_func=None, input_func=None):
        self.name = "human"
        self.description = (
            "You can ask a human for guidance when you think you "
            "got stuck or you are not sure what to do next. "
            "The input should be a question for the human."
        )
        self.prompt_func = prompt_func if prompt_func else self._print_func
        self.input_func = input_func if input_func else input

    def _print_func(self, text: str) -> None:
        print("\n")
        print(text)

    def run(self, query: str) -> str:
        """Use the Human input tool."""
        self.prompt_func(query)
        return self.input_func()


def generate_agent_role_prompt(agent):
    """ Generates the agent role prompt.
    Args: agent (str): The type of the agent.
    Returns: str: The agent role prompt.
    """
    prompts = {
        "Finance Agent": "You are a seasoned finance analyst AI assistant. Your primary goal is to compose comprehensive, astute, impartial, and methodically arranged financial reports based on provided data and trends.",
        "Travel Agent": "You are a world-travelled AI tour guide assistant. Your main purpose is to draft engaging, insightful, unbiased, and well-structured travel reports on given locations, including history, attractions, and cultural insights.",
        "Academic Research Agent": "You are an AI academic research assistant. Your primary responsibility is to create thorough, academically rigorous, unbiased, and systematically organized reports on a given research topic, following the standards of scholarly work.",
        "Default Agent": "You are an AI critical thinker research assistant. Your sole purpose is to write well written, critically acclaimed, objective and structured reports on given text."

    }

    return prompts.get(agent, "No such agent")


def generate_report_prompt(question, research_summary):
    """ Generates the report prompt for the given question and research summary.
    Args: question (str): The question to generate the report prompt for
            research_summary (str): The research summary to generate the report prompt for
    Returns: str: The report prompt for the given question and research summary
    """

    return f'"""{research_summary}""" Using the above information, answer the following'\
           f' question or topic: "{question}" in a detailed report --'\
           " The report should focus on the answer to the question, should be well structured, informative," \
           " in depth, with facts and numbers if available, a minimum of 1,200 words and with markdown syntax and apa format. "\
            "Write all source urls at the end of the report in apa format"

def generate_search_queries_prompt(question):
    """ Generates the search queries prompt for the given question.
    Args: question (str): The question to generate the search queries prompt for
    Returns: str: The search queries prompt for the given question
    """

    return f'Write 4 google search queries to search online that form an objective opinion from the following: "{question}"'\
           f'You must respond with a list of strings in the following format: ["query 1", "query 2", "query 3", "query 4"]'


def generate_resource_report_prompt(question, research_summary):
    """Generates the resource report prompt for the given question and research summary.

    Args:
        question (str): The question to generate the resource report prompt for.
        research_summary (str): The research summary to generate the resource report prompt for.

    Returns:
        str: The resource report prompt for the given question and research summary.
    """
    return f'"""{research_summary}""" Based on the above information, generate a bibliography recommendation report for the following' \
           f' question or topic: "{question}". The report should provide a detailed analysis of each recommended resource,' \
           ' explaining how each source can contribute to finding answers to the research question.' \
           ' Focus on the relevance, reliability, and significance of each source.' \
           ' Ensure that the report is well-structured, informative, in-depth, and follows Markdown syntax.' \
           ' Include relevant facts, figures, and numbers whenever available.' \
           ' The report should have a minimum length of 1,200 words.'


def generate_outline_report_prompt(question, research_summary):
    """ Generates the outline report prompt for the given question and research summary.
    Args: question (str): The question to generate the outline report prompt for
            research_summary (str): The research summary to generate the outline report prompt for
    Returns: str: The outline report prompt for the given question and research summary
    """

    return f'"""{research_summary}""" Using the above information, generate an outline for a research report in Markdown syntax'\
           f' for the following question or topic: "{question}". The outline should provide a well-structured framework'\
           ' for the research report, including the main sections, subsections, and key points to be covered.' \
           ' The research report should be detailed, informative, in-depth, and a minimum of 1,200 words.' \
           ' Use appropriate Markdown syntax to format the outline and ensure readability.'

def generate_concepts_prompt(question, research_summary):
    """ Generates the concepts prompt for the given question.
    Args: question (str): The question to generate the concepts prompt for
            research_summary (str): The research summary to generate the concepts prompt for
    Returns: str: The concepts prompt for the given question
    """

    return f'"""{research_summary}""" Using the above information, generate a list of 5 main concepts to learn for a research report'\
           f' on the following question or topic: "{question}". The outline should provide a well-structured framework'\
           'You must respond with a list of strings in the following format: ["concepts 1", "concepts 2", "concepts 3", "concepts 4, concepts 5"]'


def generate_lesson_prompt(concept):
    """
    Generates the lesson prompt for the given question.
    Args:
        concept (str): The concept to generate the lesson prompt for.
    Returns:
        str: The lesson prompt for the given concept.
    """

    prompt = f'generate a comprehensive lesson about {concept} in Markdown syntax. This should include the definition'\
    f'of {concept}, its historical background and development, its applications or uses in different'\
    f'fields, and notable events or facts related to {concept}.'

    return prompt

def get_report_by_type(report_type):
    report_type_mapping = {
        'research_report': generate_report_prompt,
        'resource_report': generate_resource_report_prompt,
        'outline_report': generate_outline_report_prompt
    }
    return report_type_mapping[report_type]


import torch
import logging
from transformers import AutoModelForCausalLM, AutoTokenizer, BitsAndBytesConfig

class HuggingFaceLLM:
    def __init__(self, model_id: str, device: str = None, max_length: int = 20, quantize: bool = False, quantization_config: dict = None):
        self.logger = logging.getLogger(__name__)
        self.device = device if device else ('cuda' if torch.cuda.is_available() else 'cpu')
        self.model_id = model_id
        self.max_length = max_length

        bnb_config = None
        if quantize:
            if not quantization_config:
                quantization_config = {
                    'load_in_4bit': True,
                    'bnb_4bit_use_double_quant': True,
                    'bnb_4bit_quant_type': "nf4",
                    'bnb_4bit_compute_dtype': torch.bfloat16
                }
            bnb_config = BitsAndBytesConfig(**quantization_config)

        try:
            self.tokenizer = AutoTokenizer.from_pretrained(self.model_id)
            self.model = AutoModelForCausalLM.from_pretrained(self.model_id, quantization_config=bnb_config)
            self.model.to(self.device)
        except Exception as e:
            self.logger.error(f"Failed to load the model or the tokenizer: {e}")
            raise

    def __call__(self, prompt_text: str, max_length: int = None):
        max_length = max_length if max_length else self.max_length
        try:
            inputs = self.tokenizer.encode(prompt_text, return_tensors="pt").to(self.device)
            with torch.no_grad():
                outputs = self.model.generate(inputs, max_length=max_length, do_sample=True)
            return self.tokenizer.decode(outputs[0], skip_special_tokens=True)
        except Exception as e:
            self.logger.error(f"Failed to generate the text: {e}")
            raise


    def generate(self, prompt_text: str, max_length: int = None):
        max_length = max_length if max_length else self.max_length
        try:
            inputs = self.tokenizer.encode(prompt_text, return_tensors="pt").to(self.device)
            with torch.no_grad():
                outputs = self.model.generate(inputs, max_length=max_length, do_sample=True)
            return self.tokenizer.decode(outputs[0], skip_special_tokens=True)
        except Exception as e:
            self.logger.error(f"Failed to generate the text: {e}")
            raise



"""Agent Infrastructure, models, memory, utils, tools"""


#models
from swarms.agents.models.anthropic import Anthropic
from swarms.agents.models.huggingface import HuggingFaceLLM
from swarms.agents.models.palm import GooglePalm
from swarms.agents.models.petals import Petals
from swarms.agents.models.openai import OpenAI

###########
#tools
from swarms.agents.tools.base import BaseTool, Tool, StructuredTool, ToolWrapper, BaseToolSet, ToolCreator, GlobalToolsCreator, SessionToolsCreator, ToolsFactory
from swarms.agents.tools.autogpt import pushd, process_csv, async_load_playwright, run_async, browse_web_page, WebpageQATool, web_search, query_website_tool
from swarms.agents.tools.exit_conversation import ExitConversation

from swarms.agents.tools.models import MaskFormer, ImageEditing, InstructPix2Pix, Text2Image, VisualQuestionAnswering, ImageCaptioning
from swarms.agents.tools.file_mangagement import read_tool, write_tool, list_tool
from swarms.agents.tools.requests import RequestsGet

from swarms.agents.tools.developer import Terminal, CodeEditor



import asyncio
import logging
from typing import Optional

from langchain import OpenAI

from swarms.boss.boss_node import BossNode
from swarms.workers.worker_node import WorkerNode

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')

# TODO: Pass in abstract LLM class that can utilize Hf or Anthropic models, Move away from OPENAI
# TODO: ADD Universal Communication Layer, a ocean vectorstore instance
# TODO: BE MORE EXPLICIT ON TOOL USE, TASK DECOMPOSITION AND TASK COMPLETETION AND ALLOCATION
# TODO: Add RLHF Data collection, ask user how the swarm is performing
# TODO: Create an onboarding process if not settings are preconfigured like `from swarms import Swarm, Swarm()` => then initiate onboarding name your swarm + provide purpose + etc 


# ---------- Constants ----------
ROOT_DIR = "./data/"


class HierarchicalSwarm:
    def __init__(
        self, 
        openai_api_key: Optional[str] = "", 
        use_vectorstore: Optional[bool] = True, 

        use_async: Optional[bool] = True, 
        worker_name: Optional[str] = "Swarm Worker AI Assistant",
        verbose: Optional[bool] = False,
        human_in_the_loop: Optional[bool] = True, 
        boss_prompt: Optional[str] = "You are an Boss in a swarm who performs one task based on the following objective: {objective}. Take into account these previously completed tasks: {context}.\n",
        worker_prompt: Optional[str] = None,
        temperature: Optional[float] = 0.5,
        max_iterations: Optional[int] = None,
        logging_enabled: Optional[bool] = True):
            
        self.openai_api_key = openai_api_key
        self.use_vectorstore = use_vectorstore
        self.use_async = use_async
        self.worker_name = worker_name
        self.human_in_the_loop = human_in_the_loop

        self.boss_prompt = boss_prompt
        self.temperature = temperature
        self.max_iterations = max_iterations
        self.logging_enabled = logging_enabled
        self.verbose = verbose

        self.worker_node = WorkerNode(
            openai_api_key=self.openai_api_key,
            worker_name=self.worker_name,
            temperature=self.temperature,
            human_in_the_loop=self.human_in_the_loop,
            verbose=self.verbose
        )

        self.boss_node = BossNode(
            api_key=self.openai_api_key,
            worker_node=self.worker_node,
            llm_class=OpenAI,
            max_iterations=self.max_iterations,
            verbose=self.verbose
        )

        self.logger = logging.getLogger()
        if not logging_enabled:
            self.logger.disabled = True

    def run(self, objective):
        try:
            self.boss_node.task = self.boss_node.create_task(objective)
            logging.info(f"Running task: {self.boss_node.task}")
            if self.use_async:
                loop = asyncio.get_event_loop()
                result = loop.run_until_complete(self.boss_node.run())
            else:
                result = self.boss_node.run()
            logging.info(f"Completed tasks: {self.boss_node.task}")
            return result
        except Exception as e:
            logging.error(f"An error occurred in run: {e}")
            return None

def swarm(
    api_key: Optional[str]="", 
    objective: Optional[str]="", 
    ):

    if not api_key or not isinstance(api_key, str):
        logging.error("Invalid OpenAI key")
        raise ValueError("A valid OpenAI API key is required")
    if not objective or not isinstance(objective, str):
        logging.error("Invalid objective")
        raise ValueError("A valid objective is required")
    try:
        swarms = HierarchicalSwarm(api_key, use_async=False) #logging_enabled=logging_enabled) # Turn off async
        result = swarms.run(objective)
        if result is None:
            logging.error("Failed to run swarms")
        else:
            logging.info(f"Successfully ran swarms with results: {result}")
        return result
    except Exception as e:
        logging.error(f"An error occured in swarm: {e}")
        return None


import requests
import os

class Anthropic:
    """Anthropic large language models."""

    def __init__(self, model="claude-2", max_tokens_to_sample=256, temperature=None, top_k=None, top_p=None, streaming=False, default_request_timeout=None):
        self.model = model
        self.max_tokens_to_sample = max_tokens_to_sample
        self.temperature = temperature
        self.top_k = top_k
        self.top_p = top_p
        self.streaming = streaming
        self.default_request_timeout = default_request_timeout or 600
        self.anthropic_api_url = os.getenv("ANTHROPIC_API_URL", "https://api.anthropic.com")
        self.anthropic_api_key = os.getenv("ANTHROPIC_API_KEY")

    def _default_params(self):
        """Get the default parameters for calling Anthropic API."""
        d = {
            "max_tokens_to_sample": self.max_tokens_to_sample,
            "model": self.model,
        }
        if self.temperature is not None:
            d["temperature"] = self.temperature
        if self.top_k is not None:
            d["top_k"] = self.top_k
        if self.top_p is not None:
            d["top_p"] = self.top_p
        return d

    def generate(self, prompt, stop=None):
        """Call out to Anthropic's completion endpoint."""
        stop = stop or []
        params = self._default_params()
        headers = {"Authorization": f"Bearer {self.anthropic_api_key}"}
        data = {
            "prompt": prompt,
            "stop_sequences": stop,
            **params
        }
        response = requests.post(f"{self.anthropic_api_url}/completions", headers=headers, json=data, timeout=self.default_request_timeout)
        return response.json().get("completion")
    
    def __call__(self, prompt, stop=None):
        """Call out to Anthropic's completion endpoint."""
        stop = stop or []
        params = self._default_params()
        headers = {"Authorization": f"Bearer {self.anthropic_api_key}"}
        data = {
            "prompt": prompt,
            "stop_sequences": stop,
            **params
        }
        response = requests.post(f"{self.anthropic_api_url}/completions", headers=headers, json=data, timeout=self.default_request_timeout)
        return response.json().get("completion")


#input agent or multiple: => it handles multi agent communication, it handles task assignment, task execution, report back with a status, auto scaling,  number of agent nodes,
#make it optional to have distributed communication protocols, trco, rdp, http, microsoervice, make it optional to collect data from users runs
"""
# Orchestrator
* Takes in an agent class with vector store, then handles all the communication and scales up a swarm with number of agents and handles task assignment and task completion

```python

from swarms import OpenAI, Orchestrator, Swarm

orchestrated = Orchestrate(OpenAI, nodes=40) #handles all the task assignment and allocation and agent communication using a vectorstore as a universal communication layer and also handlles the task completion logic

Objective = "Make a business website for a marketing consultancy"

Swarms = (Swarms(orchestrated, auto=True, Objective))
```





In terms of architecture, the swarm might look something like this:

```
                                           (Orchestrator)
                                             /        \
            ( Agent)---(Communication Layer)       (Communication Layer)---(Agent)
              /                  |                                           |                 \
(Task Assignment)      (Task Completion)                    (Task Assignment)       (Task Completion)
```

Each LLM agent communicates with the orchestrator through a dedicated communication layer. The orchestrator assigns tasks to each LLM agent, which the agents then complete and return. This setup allows for a high degree of flexibility, scalability, and robustness.

In the context of swarm LLMs, one could consider an **Omni-Vector Embedding Database** for communication. This database could store and manage the high-dimensional vectors produced by each LLM agent.

- Strengths: This approach would allow for similarity-based lookup and matching of LLM-generated vectors, which can be particularly useful for tasks that involve finding similar outputs or recognizing patterns.

- Weaknesses: An Omni-Vector Embedding Database might add complexity to the system in terms of setup and maintenance. It might also require significant computational resources, depending on the volume of data being handled and the complexity of the vectors. The handling and transmission of high-dimensional vectors could also pose challenges in terms of network load.


from swarms import WorkerNode, Orchestrate

Orchestrate(WorkerNode, autoscale=True, nodes=int, swarm_type="flat")

"""
#in a shared state, provide 3 communication times, during task assignment, task compeltion, and feedback or inability to complete a task.
import threading
from abc import ABC, abstractmethod
from typing import Any, Dict, List

from swarms.agents.memory.ocean import OceanDB


class Orchestrator(ABC):
    def __init__(self, 
    agent, 
    agent_list: List[Any], 
    task_queue: List[Any], 
    vector_db: OceanDB
    ):
        self.agent = agent
        self.agents = [agent_class() for _ in range(agent_list)]
        self.task_queue = task_queue
        self.vector_db = vector_db
        self.current_tasks = {}
        self.lock = threading.Lock()
        
    @abstractmethod
    def assign_task(self, agent_id: int, task: Dict[str, Any]) -> None:
        """Assign a task to a specific agent"""
        with self.lock:
            if self.task_queue:
                #get and agent and a task
                agent = self.agents.pop(0)
                task = self.task_queue.popleft()

                #process the task and get result and vector representation
                result, vector_representation = agent.process_task()

                #store the vector representation in the database
                self.vector_db.add_documents([vector_representation],[str(id(task))])

                #put the agent back to agent slist
                self.agents.append(agent)

                logging.info(f"Task {id(str)} has been processed by agent {id(agent)} ")

                return result
            else:
                logging.error("Task queue is empty")
    
    @abstractmethod
    def retrieve_results(self, agent_id: int) -> Any:
        """Retrieve results from a specific agent"""
        try:
            #Query the vector database for documents created by the agents 
            results = self.vector_db.query(query_texts=[str(agent_id)], n_results=10)
            return results
        except Exception as e:
            logging.error(f"Failed to retrieve results from agent {agent_id}. Error {e}")
            raise
    
    @abstractmethod
    def update_vector_db(self, data) -> None:
        """Update the vector database"""
        try:
            self.vector_db.add_documents([data['vector']], [str(data['task_id'])])
        except Exception as e:
            logging.error(f"Failed to update the vector database. Error: {e}")
            raise


    @abstractmethod
    def get_vector_db(self):
        """Retrieve the vector database"""
        return self.vector_db

    def append_to_db(self, collection, result: str):
        """append the result of the swarm to a specifici collection in the database"""
        try:
            self.vector_db.append_document(collection, result, id=str(id(result)))
        except Exception as e:
            logging.error(f"Failed to append the agent output to database. Error: {e}")
            raise

    def run(self, objective:str, collection):
        """Runs"""

        if not objective or not isinstance(objective, str):
            logging.error("Invalid objective")
            raise ValueError("A valid objective is required")
        
        try:
            #add objective to agent
            self.task_queue.append(objective)

            #assign tasks to agents
            results = [self.assign_task(agent_id, task) for agent_id, task in zip(range(len(self.agents)), self.task_queue)]
            
            for result in results:
                self.append_to_db(collection, result)
            

            logging.info(f"Successfully ran swarms with results: {results}")
            return results
        except Exception as e:
            logging.error(f"An error occured in swarm: {e}")
            return None




#PRE CONFIGURED AGENTS WITH domain explicit TOOLS
#Build your own Agent
# Learn from previous runs in session management => it's a sucessful run => omniversal memory for all swarms  


import logging
from typing import Union
from pegasus import Pegasus

# import oceandb
# from oceandb.utils.embedding_functions import MultiModalEmbeddingfunction


class PegasusEmbedding:
    def __init__(self, modality: str, multi_process: bool = False, n_processes: int = 4):
        self.modality = modality
        self.multi_process = multi_process
        self.n_processes = n_processes
        try:
            self.pegasus = Pegasus(modality, multi_process, n_processes)
        except Exception as e:
            logging.error(f"Failed to initialize Pegasus with modality: {modality}: {e}")
            raise
    
    def embed(self, data: Union[str, list[str]]):
        try:
            return self.pegasus.embed(data)
        except Exception as e:
            logging.error(f"Failed to generate embeddings. Error: {e}")
            raise




from abc import ABC, abstractmethod

class AbstractModel(ABC):
    #abstract base class for language models
    @abstractmethod
    def generate(self, prompt):
        #generate text using language model
        pass
    


#kye 
#aug 8, 11:51

from simpleaichat import AIChat, AsyncAIChat
import asyncio


class OpenAI:
    def __init__(self,
                 api_key=None,
                 system=None,
                 console=True,
                 model=None,
                 params=None,
                 save_messages=True):
        self.api_key = api_key or self._fetch_api_key()
        self.system = system or "You are a helpful assistant"

        try:
                
            self.ai = AIChat(api_key=self.api_key, 
                system=self.system, 
                console=console, 
                model=model, 
                params=params, 
                save_messages=save_messages)
            
            self.async_ai = AsyncAIChat(
                api_key=self.api_key,
                system=self.system,
                console=console,
                model=model,
                params=params,
                save_messages=save_messages
            )
        
        except Exception as error:
            raise ValueError(f"Failed to initialize the chat with error: {error}, check inputs and input types")
    
    def __call__(self, message, **kwargs):
        try:
            return self.ai(message, **kwargs)
        except Exception as error:
            print(f"Error in OpenAI, {error}")
    
    def generate(self, message, **kwargs):
        try:
            return self.ai(message, **kwargs)
        except Exception as error:
            print(f"Error in OpenAI, {error}")
        
    async def generate_async(self, message, **kwargs):
        try:
            return await self.async_ai(message, **kwargs)
        except Exception as error:
            raise Exception(f"Error in asynchronous OpenAI Call, {error}")
    
    def initialize_chat(self, ids):
        for id in ids:
            try:
                self.async_ai.new_session(api_key=self.api_key, id=id)
            except Exception as error:
                raise ValueError(f"Failed to initialize session for ID {id} with error: {error}")
    
    async def ask_multiple(self, ids, question_template):
        try:
            self.initialize_chat(ids)
            tasks = [self.async_ai(question_template.format(id=id), id=id) for id in ids]
            return await asyncio.gather(*tasks)
        except Exception as error:
            raise Exception(f"Error in ask_multiple: method: {error}")
    
    async def stream_multiple(self, ids, question_template):
        try:
            self.initialize_chat(ids)

            async def stream_id(id):
                async for chunk in await self.async_ai.stream(question_template.format(id=id), id=id):
                    response = chunk["response"]
                return response
        
            tasks = [stream_id(id) for id in ids]
            return await asyncio.gather(*tasks)
        except Exception as error:
            raise Exception(f"Error in stream_multiple method: {error}")
    
    def fetch_api_key(self):
        pass


#usage
#from swarms import OpenAI()
#chat = OpenAI()
#response = chat.generate("Hello World")
#print(response)

#async
# async_responses = asyncio.run(chat.ask_multiple(['id1', 'id2'], "How is {id}"))
# print(async_responses)



from swarms.agents.tools.base import BaseToolSet, SessionGetter, ToolScope, tool
from swarms.utils.logger import logger


class ExitConversation(BaseToolSet):
    @tool(
        name="Exit Conversation",
        description="A tool to exit the conversation. "
        "Use this when you want to exit the conversation. "
        "The input should be a message that the conversation is over.",
        scope=ToolScope.SESSION,
    )
    def exit(self, message: str, get_session: SessionGetter) -> str:
        """Run the tool."""
        _, executor = get_session()
        del executor

        logger.debug("\nProcessed ExitConversation.")

        return message
    





from swarms.swarms.swarms import HierarchicalSwarm, swarm
from swarms.workers.worker_node import WorkerNodeInitializer, WorkerNode, worker_node
from swarms.boss.boss_node import BossNodeInitializer, BossNode

#models
from swarms.agents.models.anthropic import Anthropic
from swarms.agents.models.huggingface import HuggingFaceLLM
from swarms.agents.models.palm import GooglePalm
from swarms.agents.models.petals import Petals
from swarms.agents.models.openai import OpenAI



import asyncio
import os

# Tools
from contextlib import contextmanager
from typing import Optional

import pandas as pd
from langchain.agents import tool
from langchain.agents.agent_toolkits.pandas.base import create_pandas_dataframe_agent
from langchain.chains.qa_with_sources.loading import load_qa_with_sources_chain
from langchain.docstore.document import Document
from langchain.memory.chat_message_histories import FileChatMessageHistory
from langchain.tools.human.tool import HumanInputRun

ROOT_DIR = "./data/"

from langchain.chains.qa_with_sources.loading import BaseCombineDocumentsChain
from langchain.chat_models import ChatOpenAI
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain.tools import BaseTool, DuckDuckGoSearchRun
from langchain.tools.file_management.read import ReadFileTool
from langchain.tools.file_management.write import WriteFileTool
from pydantic import Field

llm = ChatOpenAI(model_name="gpt-4", temperature=1.0)




@contextmanager
def pushd(new_dir):
    """Context manager for changing the current working directory."""
    prev_dir = os.getcwd()
    os.chdir(new_dir)
    try:
        yield
    finally:
        os.chdir(prev_dir)

@tool
def process_csv(
    llm, csv_file_path: str, instructions: str, output_path: Optional[str] = None
) -> str:
    """Process a CSV by with pandas in a limited REPL.\
 Only use this after writing data to disk as a csv file.\
 Any figures must be saved to disk to be viewed by the human.\
 Instructions should be written in natural language, not code. Assume the dataframe is already loaded."""
    with pushd(ROOT_DIR):
        try:
            df = pd.read_csv(csv_file_path)
        except Exception as e:
            return f"Error: {e}"
        agent = create_pandas_dataframe_agent(llm, df, max_iterations=30, verbose=False)
        if output_path is not None:
            instructions += f" Save output to disk at {output_path}"
        try:
            result = agent.run(instructions)
            return result
        except Exception as e:
            return f"Error: {e}"
        

async def async_load_playwright(url: str) -> str:
    """Load the specified URLs using Playwright and parse using BeautifulSoup."""
    from bs4 import BeautifulSoup
    from playwright.async_api import async_playwright

    results = ""
    async with async_playwright() as p:
        browser = await p.chromium.launch(headless=True)
        try:
            page = await browser.new_page()
            await page.goto(url)

            page_source = await page.content()
            soup = BeautifulSoup(page_source, "html.parser")

            for script in soup(["script", "style"]):
                script.extract()

            text = soup.get_text()
            lines = (line.strip() for line in text.splitlines())
            chunks = (phrase.strip() for line in lines for phrase in line.split("  "))
            results = "\n".join(chunk for chunk in chunks if chunk)
        except Exception as e:
            results = f"Error: {e}"
        await browser.close()
    return results

def run_async(coro):
    event_loop = asyncio.get_event_loop()
    return event_loop.run_until_complete(coro)

@tool
def browse_web_page(url: str) -> str:
    """Verbose way to scrape a whole webpage. Likely to cause issues parsing."""
    return run_async(async_load_playwright(url))


def _get_text_splitter():
    return RecursiveCharacterTextSplitter(
        # Set a really small chunk size, just to show.
        chunk_size = 500,
        chunk_overlap  = 20,
        length_function = len,
    )


class WebpageQATool(BaseTool):
    name = "query_webpage"
    description = "Browse a webpage and retrieve the information relevant to the question."
    text_splitter: RecursiveCharacterTextSplitter = Field(default_factory=_get_text_splitter)
    qa_chain: BaseCombineDocumentsChain
    
    def _run(self, url: str, question: str) -> str:
        """Useful for browsing websites and scraping the text information."""
        result = browse_web_page.run(url)
        docs = [Document(page_content=result, metadata={"source": url})]
        web_docs = self.text_splitter.split_documents(docs)
        results = []
        # TODO: Handle this with a MapReduceChain
        for i in range(0, len(web_docs), 4):
            input_docs = web_docs[i:i+4]
            window_result = self.qa_chain({"input_documents": input_docs, "question": question}, return_only_outputs=True)
            results.append(f"Response from window {i} - {window_result}")
        results_docs = [Document(page_content="\n".join(results), metadata={"source": url})]
        return self.qa_chain({"input_documents": results_docs, "question": question}, return_only_outputs=True)
    
    async def _arun(self, url: str, question: str) -> str:
        raise NotImplementedError


query_website_tool = WebpageQATool(qa_chain=load_qa_with_sources_chain(llm))

# !pip install duckduckgo_search
web_search = DuckDuckGoSearchRun()




import logging
import os
from typing import Dict, List

from langchain.memory.chat_message_histories import FileChatMessageHistory


logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')




from typing import Dict, List

from langchain.memory.chat_message_histories import FileChatMessageHistory

from swarms.agents.tools.main import (
    BaseToolSet,
    CodeEditor,
    ExitConversation,
    RequestsGet,
    Terminal,
)
from swarms.utils.main import BaseHandler, CsvToDataframe, FileType


class WorkerUltraNode:
    """Useful for when you need to spawn an autonomous agent instance as a worker to accomplish complex tasks, it can search the internet or spawn child multi-modality models to process and generate images and text or audio and so on"""

    def __init__(self, llm, toolsets, vectorstore):
        if not llm or not toolsets or not vectorstore:
            logging.error("llm, toolsets, and vectorstore cannot be None.")
            raise ValueError("llm, toolsets, and vectorstore cannot be None.")
        
        self.llm = llm
        self.toolsets = toolsets
        self.vectorstore = vectorstore
        self.agent = None

    def create_agent(self, ai_name="Swarm Worker AI Assistant", ai_role="Assistant", human_in_the_loop=False, search_kwargs={}, verbose=False):
        logging.info("Creating agent in WorkerNode")
        try:
            tools_list = list(self.toolsets.values())
            self.agent = AutoGPT.from_llm_and_tools(
                ai_name=ai_name,
                ai_role=ai_role,
                tools=tools_list,  # Pass the dictionary instead of the list
                llm=self.llm,
                memory=self.vectorstore.as_retriever(search_kwargs=search_kwargs),
                human_in_the_loop=human_in_the_loop,
                chat_history_memory=FileChatMessageHistory("chat_history.txt"),
            )
            self.agent.chain.verbose = verbose
        except Exception as e:
            logging.error(f"Error while creating agent: {str(e)}")
            raise e

    def add_toolset(self, toolset: BaseToolSet):
        if not isinstance(toolset, BaseToolSet):
            logging.error("Toolset must be an instance of BaseToolSet.")
            raise TypeError("Toolset must be an instance of BaseToolSet.")
        
        self.toolsets.append(toolset)

    def run(self, prompt: str) -> str:
        if not isinstance(prompt, str):
            logging.error("Prompt must be a string.")
            raise TypeError("Prompt must be a string.")
        
        if not prompt:
            logging.error("Prompt is empty.")
            raise ValueError("Prompt is empty.")
        
        try:
            self.agent.run([f"{prompt}"])
            return "Task completed by WorkerNode"
        except Exception as e:
            logging.error(f"While running the agent: {str(e)}")
            raise e

class WorkerUltraNodeInitializer:
    def __init__(self, openai_api_key):
        if not openai_api_key:
            logging.error("OpenAI API key is not provided")
            raise ValueError("openai_api_key cannot be None")
        
        self.openai_api_key = openai_api_key

    def initialize_llm(self, llm_class, temperature=0.5):
        if not llm_class:
            logging.error("llm_class cannot be none")
            raise ValueError("llm_class cannot be None")
        
        try:
            return llm_class(openai_api_key=self.openai_api_key, temperature=temperature)
        except Exception as e:
            logging.error(f"Failed to initialize language model: {e}")
            raise

    def initialize_toolsets(self):
        try:
            toolsets: List[BaseToolSet] = [
                Terminal(),
                CodeEditor(),
                RequestsGet(),
                ExitConversation(),
            ]
            handlers: Dict[FileType, BaseHandler] = {FileType.DATAFRAME: CsvToDataframe()}

            if os.environ.get("USE_GPU", False):
                import torch

                from swarms.agents.tools.main import (
                    ImageCaptioning,
                    ImageEditing,
                    InstructPix2Pix,
                    Text2Image,
                    VisualQuestionAnswering,
                )

                if torch.cuda.is_available():
                    toolsets.extend(
                        [
                            Text2Image("cuda"),
                            ImageEditing("cuda"),
                            InstructPix2Pix("cuda"),
                            VisualQuestionAnswering("cuda"),
                        ]
                    )
                    handlers[FileType.IMAGE] = ImageCaptioning("cuda")

            return toolsets
        except Exception as e:
            logging.error(f"Failed to initialize toolsets: {e}")

    def initialize_vectorstore(self):
        try:
                
            embeddings_model = OpenAIEmbeddings(openai_api_key=self.openai_api_key)
            embedding_size = 1536
            index = faiss.IndexFlatL2(embedding_size)
            return FAISS(embeddings_model.embed_query, index, InMemoryDocstore({}), {})
        except Exception as e:
            logging.error(f"Failed to initialize vector store: {e}")
            raise

    def create_worker_node(self, llm_class=ChatOpenAI, ai_name="Swarm Worker AI Assistant", ai_role="Assistant", human_in_the_loop=False, search_kwargs={}, verbose=False):
        if not llm_class:
            logging.error("llm_class cannot be None.")
            raise ValueError("llm_class cannot be None.")
        try:
            worker_toolsets = self.initialize_toolsets()
            vectorstore = self.initialize_vectorstore()
            worker_node = WorkerUltraNode(llm=self.initialize_llm(llm_class), toolsets=worker_toolsets, vectorstore=vectorstore)
            worker_node.create_agent(ai_name=ai_name, ai_role=ai_role, human_in_the_loop=human_in_the_loop, search_kwargs=search_kwargs, verbose=verbose)
            return worker_node
        except Exception as e:
            logging.error(f"Failed to create worker node: {e}")
            raise

def worker_ultra_node(openai_api_key):
    if not openai_api_key:
        logging.error("OpenAI API key is not provided")
        raise ValueError("OpenAI API key is required")
    
    try:
        initializer = WorkerUltraNodeInitializer(openai_api_key)
        worker_node = initializer.create_worker_node()
        return worker_node
    except Exception as e:
        logging.error(f"An error occurred in worker_node: {e}")
        raise